#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <poll.h>
#include <sys/socket.h>
#include <arpa/inet.h>

enum {
  RST_WAITSTART,
  RST_DATA,
  RST_ESCDATA,
  RST_REPDATA,
  RST_CSUM1,
  RST_CSUM2
};

static int gdb_sk;

int gdb_connect( const char *host, int port ) {
  gdb_sk = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
  if ( gdb_sk == -1 ) return -1;

  sockaddr_in sin;
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = (host && *host)? inet_addr( host ):
                                         htonl(INADDR_LOOPBACK);
  sin.sin_port = htons( port );

  return connect( gdb_sk, (sockaddr*)&sin, sizeof(sin) );
}

int gdb_connect( const char *host_port ) {
  char *arg = strdup( host_port );

  char *pos = strchr( arg, ':' );
  if ( !pos ) return -1;

  *pos = 0;

  int rc =  gdb_connect( arg, atoi(pos + 1) );

  free( arg );

  return rc;
}

static int gdb_getreply( unsigned char *buf, int count ) {
  unsigned char *buf_start = buf, prev = 0, remote_csum, csum, c;
  char csum_buf[3];
  int state = RST_WAITSTART;

  csum = 0;

  for (;;) {
	int rc = read( gdb_sk, &c, 1 );
	if ( rc < 0 ) break;
	if ( !rc ) continue;

//	putchar( c );

	switch ( state ) {
	case RST_WAITSTART:
		if ( c == '$' ) state = RST_DATA;
		break;
	case RST_DATA:
		if ( c == '#' ) {
			state = RST_CSUM1;
			break;
		}
		
		csum += c;

		if ( c == '{' ) state = RST_ESCDATA;
		else if ( c == '*' ) state = RST_REPDATA;
		else if ( buf < buf_start + count )
			*buf++ = prev = c;
		break;
	case RST_ESCDATA:
		csum += c;

		if ( buf < buf_start + count )
			*buf++ = prev = c ^ 0x20;

		state = RST_DATA;
		break;
	case RST_REPDATA:
		csum += c;

		if ( c < 29 ) break;
		for ( int i = 0; i < c - 29 && buf < buf_start + count; i++ )
 			*buf++ = prev;
		break;
	case RST_CSUM1:
		csum_buf[0] = c;
		state = RST_CSUM2;
		break;
	case RST_CSUM2:
		csum_buf[1] = c;
		csum_buf[2] = 0;

		remote_csum = strtol( csum_buf, 0, 16 );
		if ( remote_csum != csum ) {
			printf( "\nchecksum mismatch received=%02x, calculated=%02x\n",
				remote_csum, csum );
			return -1;
		}
		goto done;
	}
  }

done:
  c = '+';
  write( gdb_sk, &c, 1 );

  return buf - buf_start;
}

static int gdb_waitack() {
  char c;

  pollfd ack;
  ack.fd = gdb_sk;
  ack.events = POLLIN;

  int rc = poll( &ack, 1, 500 );
  if ( rc < 1 || !(ack.revents & POLLIN) ) return -1;

  if ( read( gdb_sk, &c, 1 ) != 1 || c != '+' ) return -1;

  return 0;
}

int gdb_readmem( unsigned char *buf, int addr, int count ) {
  char cmd[30];
  unsigned char csum = 0;

  sprintf( cmd, "$m%08x,%x#", addr, count );
  for ( int i = 1; i < strlen(cmd) - 1; i++ )
  	csum += cmd[i];

  sprintf( cmd + strlen(cmd), "%02x", csum );

  write( gdb_sk, cmd, strlen(cmd) );
  if ( gdb_waitack() < 0 ) return -1;

  unsigned char *gdb_buf = (unsigned char*)malloc(count * 2 + 1);
  int rc = gdb_getreply( gdb_buf, count * 2 );

  if ( rc < 0 || (rc & 1) ) {
  	free( gdb_buf );
  	return -1;
  }

  unsigned char c;
  for ( int i = 0; i < rc; i += 2 ) {
  	c = gdb_buf[i + 2];
	gdb_buf[i + 2] = 0;
	buf[i >> 1] = strtol( (const char*)&gdb_buf[i], 0, 16 );
	gdb_buf[i + 2] = c;
  }

  free( gdb_buf );

  return rc >> 1;
}

int gdb_writemem( unsigned char *buf, int addr, int count ) {
  unsigned char csum = 0, c;
  char cmd[30], xbuf[100];

  sprintf( cmd, "$M%08x,%x:", addr, count );
  for ( int i = 1; i < strlen(cmd); i++ )
  	csum += cmd[i];
  write( gdb_sk, cmd, strlen(cmd) );

  for ( int i = 0; i < count; i += sizeof(xbuf)/2 ) {
  	int j = 0;
  	for ( ; j < sizeof(xbuf) && i + j/2 < count; j += 2 ) {
		c = buf[i + j/2];
		sprintf( xbuf + j, "%02x", c );
		csum += xbuf[j];
		csum += xbuf[j+1];
	}
	write( gdb_sk, xbuf, j );
  }

  sprintf( cmd, "#%02x", csum );

  write( gdb_sk, cmd, strlen(cmd) );
  if ( gdb_waitack() < 0 ) return -1;

  gdb_getreply( 0, 0 );

  return 0;
}

#if 0
void request_memory_map() {
  char cmd[50];
  unsigned char csum = 0;

  strcpy( cmd, "$qXfer:memory-map:read::0,fff" );
  for ( int i = 1; cmd[i]; i++ ) csum += cmd[i];
  sprintf( cmd + strlen(cmd), "#%02x", csum );

  write( gdb_sk, cmd, strlen(cmd) );
  if ( gdb_waitack() < 0 ) return;

  gdb_getreply( 0, 0 );
}
#endif
